import os
import sys
import threading
import signal
import socket
import logging
from swicharoo import accept_swicharoo, swicharoo_send, swicharoo_recv
from Crypto.PublicKey import RSA
import datetime
from struct import pack, unpack
from configuration import Configuration
from socket_communication import SecureCommunicationListen, SecureCommunication
import database_query as dq
import updater as ud
import rsa_helper as rh
import json


"""
the listener serves encrypted data to anyone knowing the name of the data
It stores data in a SQLite database
"""

# Set global flogging:
logging.basicConfig(
     filename='the.log',
     level=logging.INFO,
     format='[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
     datefmt='%H:%M:%S'
)
# Set up logging to console
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')  # set simple console format
console.setFormatter(formatter)
logging.getLogger('').addHandler(console)  # add the handler to the root logger

logger = logging.getLogger(__name__)
# Logging setup done

# Global shut down boolean
force_shutdown = False


class Listen:
    def __init__(self):
        self.listensocket = None
        self.clientsocket = None

        config = Configuration()
        self.ip = config.get('ip')
        self.port = config.get('port')
        self.unsecure_nick_name = config.get('unsecure_nick_name')
        self.LISTEN_READ_BYTES = int(config.get('LISTEN_READ_BYTES'))
        self.MAX_SIMULTANIOUS_CONNECTIONS = int(config.get('MAX_SIMULTANEOUS_CONNECTIONS'))
        self.listen_timeout = int(config.get('LISTEN_TIMEOUT_SECONDS'))

    def forcestop(self):
        logging.info("Shutting down listener...")
        if self.listensocket is not None:
            self.listensocket.close()

    def listenthread_secure(self, client_socket, client_address):
        logging.info("Recv Message from" + str(client_address[0]))

        prv, pub = dq.get_rsa_hex_keys()
        logging.info("Recv Message, make secure link:")
        scl = SecureCommunicationListen(client_socket, RSA.importKey(bytes.fromhex(prv)), self.LISTEN_READ_BYTES)
        if not scl.is_active():
            print('Listener could not setup secure link with caller, bailing out...')
            return
        print('Secure link created')
        n = scl.recv()

        logging.info("Recv Message done.")
        if n:
            logging.info('Received msg, size=' + str(len(n)))
        else:
            logging.info('Recved nothing, close socket and bail out.')
            return
        #logging.info('10-part:' + n[:10])

        # Allr equests take up 10 bytes
        request = n[:10]
        data = n[10:]
        #logging.info('REQ:',request)
        #print('QUERY10:', request, type(request))
        # print('MSG:',data)
        try:
            # What does he want?

            if request == b'AREUALIVE?':  # Just check if we are a Tenfingers node
                logging.info('Listen thread got AREUALIVE req.')
                scl.send(b'YES')

            # The TenFingers but in a smaller version, for now.
            if request == b'ONEFINGER:':  # Rest of the request is just a name. If the name corresponds in the database, send back the encrypted file.
                logging.info('Listen thread got ONEFINGER req.')
                data = data.decode()
                datalist = data.split(',')  # todo: This should just be a name, don't forget to escape it!

                if len(datalist) == 2:
                    print('datalist:', datalist)
                    name = datalist[0]
                    rsa_pub = datalist[1]
                    logging.info('Request for: [' + name + ']')

                    # TODO, Serve the name or an error!

                    row = dq.get_link_data_by_name_and_rsapub(name, rsa_pub)

                    if row is not None:
                        logging.info('Found [' + name + '], points on file [' + row[0] + ']')
                        # Get data from DB
                        # for now fake it and just send back
                        # fake_file="test.aes"  # fake data
                        file = row[0]
                        data = None

                        # get the folder that is in the /data/ folder (based on the table 'folder' / or our own local folder 'local'):
                        filefolder = 'data/' + dq.get_folder(rsa_pub)

                        logging.info('Trying to serve from folder:' + filefolder)

                        with open(filefolder + "/" + file + '.aes', 'rb') as f:
                            data = f.read()
                        if data is None:
                            logging.info('[' + name + '] does not point on a file or it was not readable.')
                            scl.send(b'FileNotFound')
                        else:
                            logging.info('[' + name + '] corresponds to a file and we loaded it, lets send it back:')

                            # Send it back (maybe we need to convert the data, it is raw bytes now)
                            scl.send(data)
                    else:
                        logging.info('[' + name + '] not found in database.' + rsa_pub)
                        scl.send(b'NotFound')
                else:
                    logging.info('Malformed request.')
                    scl.send(b'BadRequest')

            # Update request

            if request == b'UPDATEDATA':  # another node asks us to update data we store for them
                logging.info('Listen thread got UPDATEDATA req.')

                filesize = unpack('>Q', data[:8])[0]  # todo: always use data[:8], so a ',' in the pack(...) won't break things! SO fix it everywhere
                print('filesize', filesize)

                arr = data[8:].split(b',')  # todo: always use data[8:], so a ',' in the pack(...) won't break things! SO fix it everywhere

                other_name = arr[1].decode()
                other_pub_key_bin_hex = arr[2]

                print('Verify connecting node...')

                # First verify the caller is actually the owner of this data:
                random_checksum = b'to_be_randomised_later'  # todo:
                print('Send random verification data:', random_checksum)
                verification_digest = scl.send(random_checksum)  # caller should now have made a digest off of this checksum
                print('Got verification digest:', verification_digest)

                print('Import other pub key:', other_pub_key_bin_hex[:80], '...')

                other_pub_key = RSA.importKey(bytes.fromhex(other_pub_key_bin_hex.decode()))
                # print('Key type:', type(other_pub_key))

                print('Check digest:')
                digest_ok = rh.verify_digest(random_checksum, other_pub_key, verification_digest)

                if digest_ok:
                    print('Digest OK, lets check data size:', filesize)
                    print('other_name type:', type(other_name))
                    logging.info('Got update request for file '+other_name+' of size:'+str(filesize))

                    if filesize > 15 * 1024 * 1024:  # todo:
                        scl.send(b'Too Big!')
                    else:
                        print('File size OK.')
                        # todo: verify we actually store this file for this node!

                        other_pub_hex = other_pub_key_bin_hex.decode()
                        print('other_pub_hex:', other_pub_hex[:32], '...')
                        print('type:', type(other_pub_hex))
                        if dq.get_link_data_by_name_and_rsapub(other_name, other_pub_hex):
                            print('Update request: all good!')
                            new_file_data = scl.send(b'ACCEPTING!')

                            # Save other nodes data
                            filefolder = 'data/' + dq.get_folder(other_pub_hex) + '/'
                            if not os.path.exists(filefolder):
                                os.makedirs(filefolder)

                            save_name = filefolder + other_name + '.aes'
                            logging.info('Save off other file data: ' + save_name)
                            with open(save_name, 'wb') as f:
                                f.write(new_file_data)

                            scl.send(b'OK')
                        else:
                            print('Update data: we do not store that data here')
                            scl.send(b'DataRemoved')

                else:
                    scl.send(b'UNDIGESTABLE')
                    print('Update data: digest verification failed')

            # The classic swicharoo:
            if request == b'SWICHAROO?':  # The good old swicharoo
                logging.info('Listen thread got SWICHAROO req.')
                # We just got a request for swicharoo with a file this size:
                filesize = unpack('>Q', data)[0]

                logging.info('Swicharoo filesize request: ' + str(filesize))

                # Find the data we want to send back
                dataid = dq.get_dataid_needing_finger(filesize)
                if dataid > 0:
                    my_name, my_origfilename = dq.get_link_data_by_dataid(dataid)
                    # Send back confirmation
                    #logging.info('Send back YES_PLEASE + filesize=' + str(filesize))
                    query = b'YES_PLEASE' + pack('>Q', filesize)
                    #logging.info('send:')
                    #logging.info(query)
                    tmp = scl.send(query)
                    # todo: check we didn't get back None here!
                    tmp = tmp.decode()
                    arr = tmp.split(',')
                    logging.info('Got address data: len=' + str(len(arr)))
                    other_ip = arr[0]
                    other_port = arr[1]
                    other_pub_key = arr[2]
                    #logging.info('Got other nodes data:')
                    #logging.info(other_ip)
                    #logging.info(other_port)
                    #logging.info(other_pub_key[:10]+'...'+other_pub_key[-10:])
                    #logging.info(type(other_ip))
                    #logging.info(type(other_port))
                    #logging.info(type(other_pub_key[10:]))
                    #logging.info('Check if we know this node')
                    other_nodeid = dq.get_nodeid(other_ip, other_port, other_pub_key)
                    #logging.info('DO we know this node (0 = no)? : ' + str(other_nodeid))
                    if other_nodeid == 0:
                        print('Insert new node:', other_ip, other_port, other_pub_key[:10], '...', other_pub_key[-10:])
                        dq.insert_node(other_ip, other_port, other_pub_key)
                        other_nodeid = dq.get_nodeid(other_ip, other_port, other_pub_key)

                    if scl.is_active():  # other side didn't just close it down / reject
                        logging.info('Node connection still active, open our data and do the swicharoo')
                        # send back our data and recv others data
                        logging.info('data/local/' + my_name + '.aes')
                        filedata = None
                        with open('data/local/' + my_name + '.aes', 'rb') as f:
                            filedata = f.read()
                            logging.info('Local fie read, ready to send it.')
                        if filedata is None:
                            print('File not found', my_origfilename)
                            scl.close()
                            return

                        # recv their data
                        #logging.info('File Swicharoo...')
                        other_data = scl.send(filedata)

                        # Send our information (name, originalfilename)

                        #logging.info('Name Swicharoo...')
                        other_name = scl.send(my_name).decode()
                        logging.info('OriginalFileName Swicharoo...')
                        other_originalfilename = scl.send(my_origfilename)
                        #logging.info('ACK')
                        scl.send(b'OK')
                        #logging.info('ACK Sent')

                        # Save other nodes data
                        filefolder = 'data/' + dq.get_folder(other_pub_key) + '/'
                        if not os.path.exists(filefolder):
                            os.makedirs(filefolder)

                        save_name = filefolder + other_name + '.aes'
                        #logging.info('Save off other file data: ' + save_name)
                        with open(save_name, 'wb') as f:
                            f.write(other_data)

                        # I now have other: pubkey, name, origname, data, save off data and update DB
                        dq.insert_data_shared(other_name, other_originalfilename, other_pub_key)  # Others data is now stored here for retrieval (by name + pubkey)

                        # make a new finger with their nodeid and our dataid
                        # Inser new finger for My data
                        dq.insert_finger(other_nodeid, dataid)

                        # todo: make place for address-list in the data table / make another table for foreign node data addresses

                        # send our information ( [Address list, name, filename, timestamp], digest)
                        #ud.requre_update_data_information(scl, dataid)
                        logging.info('Swicharoo done successfully')
                    else:
                        print('Other side preemptively shut down the connection, bail out of swicharoo.')

                else:
                    logging.info('Listen thread got SWICHAROO req but WE Do not have anything to Share, so deny request!')
                    scl.send(b'NO_CARAMBA')

            if request == b'SEND_NODES':  # send back a random node for now  # todo: send back several
                # check if we already have this node, and if not, add it:
                data = data.decode()
                datalist = data.split(',')  # This should be IP/PORT data from caller todo: Escape it!
                if len(datalist) == 2:
                    port = datalist[0]
                    rsa_pub = datalist[1]

                    # This call will silently fail if the node already exists
                    dq.insert_node(client_address[0], port, rsa_pub)

                else:
                    logging.info('#Warning: SEND_NODES req malformed, len(datalist) =' + str(len(datalist)))


                nodeid = dq.get_random_node_id(False)  # todo: remove the False from the call so we send back only valid nodes
                if nodeid is not None:
                    tmp = dq.get_node_data(nodeid)
                    j = {'ip': tmp[0], 'port': tmp[1], 'rsa_pub_key': tmp[2]}
                    print('Send back node:', tmp[0], tmp[1], tmp[2][:22], flush=True)
                    # print('j:', j)
                    try:
                        r = json.dumps(j)
                    except e:
                        print('Could not dumps the dict:', str(e))
                    # print('dumps=', r, flush=True)
                    b = r.encode()  # transform to byte object
                    # print('send:', b, flush=True)
                    scl.send(b)
                else:
                    scl.send(b'{}')
                    print('We do not have any nodes to share yet!', flush=True)
                print('SEND NODES Done.', flush=True)

            # Close the socket
            client_socket.close()

            if request == b'NEW_METADT':  # A node is requiring an update of its address list, metadata,...
                tmp = data.split(b',')  # name + ',' + pub + ',' + epoch_time + address_list, digest
                if len(tmp) == 5:
                    name = tmp[0]
                    pub = tmp[1]
                    epoch_time = tmp[2]
                    digest = tmp[3]
                    address_list = tmp[4]  # address list is not used in the digest.

                    # first we verify this public key shows the private key has signed the message:
                    if rh.verify_digest(name + ',' + pub + ',' + epoch_time, pub, digest):
                        # if so, then update the database (as data is stored by name+pub_key, this works)
                        print('NEW_METADT request: verification OK, insert meta data and address list')
                        dq.try_update_address_list(name, pub, address_list, epoch_time)
                    else:
                        print('NEW_METADT request: digest verification failed')
                else:
                    print('Got mal formed NEW_METADT request, incorrect size is:', len(tmp))


        except NameError as error:
            logging.info("[NameError] Message not understood: NameError=" + str(error))
        except:
            logging.info("[General Exception] Message not understood: " + str(sys.exc_info()[0]))
            scl.send(b'GeneralError')
        logging.info('Listen thread done.')

        if scl.is_active():
            scl.close()

    def run(self):
        print('inner thread start')
        try:
            self.listensocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.listensocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.listensocket.bind((str(socket.INADDR_ANY), self.port))
            self.listensocket.listen(self.MAX_SIMULTANIOUS_CONNECTIONS)
            self.listensocket.settimeout(self.listen_timeout)
        except socket.error as err:
            print('Failed to create listen socket: ', err)
            logging.info('Failed to create listen socket')
            exit()

        # log
        logging.info("Connected on port " + str(self.port))
        global force_shutdown
        while not force_shutdown:

            gotconnection = False
            while not force_shutdown and gotconnection is False:
                try:
                    # establish connection
                    client_socket, client_address = self.listensocket.accept()
                    # Pass through the while:
                    gotconnection = True
                    # Make a new thread that will deal with the communication:
                    if not force_shutdown:
                        # Read request:
                        client_socket.settimeout(60)
                        # Start listener
                        threading.Thread(target=self.listenthread_secure, args=(client_socket, client_address)).start()

                except socket.error:  # as err:
                    # print("listen exception", err)
                    pass


print('Tenfingers listener startup: V1.02', datetime.datetime.now())

# Create thread and start up:
listener = Listen()
listener_thread = threading.Thread(name='Listener', target=listener.run)  # l.run2, args=("abc",))

logging.info("start listener thread:")
listener_thread.start()


# Set up a function that will be called at SIGINT so that a CTRL+C will shut down all threads and exit correctly:
def signal_handler(sig, frame):
        logging.info('Received SIGINT, shutting down threads.')
        global force_shutdown
        force_shutdown = True
        listener.forcestop()


# Set up the SIGINT handler
signal.signal(signal.SIGINT, signal_handler)
